import pathlib
import os
import platform
import sqlite3


class Cacher(object):
	def __init__(self):
		make_new = False
		if not self.cache_db_path.exists():
			make_new = True

		self.connection = sqlite3.connect(str(self.cache_db_path))
		if make_new:
			self.create_tables()

	@property
	def cache_folder(self):
		home_folder = pathlib.Path.home()
		cache_folder = home_folder
		if platform.system() == "Windows":
			cache_folder = cache_folder / "AppData" / "Local" / ".openet_client"
		else:
			cache_folder = cache_folder /".openet_client"

		if not cache_folder.exists():
			os.makedirs(str(cache_folder))

		return cache_folder

	@property
	def cache_db_path(self):
		return self.cache_folder / "openet_client_cache.db"

	def create_tables(self):
		cursor = self.connection.cursor()
		cursor.execute("CREATE TABLE geodatabase (location text NOT NULL UNIQUE, openet_id text)")
		self.connection.commit()
		cursor.close()

	def cache_gdb_item(self, key, value):
		try:
			cursor = self.connection.cursor()
			cursor.execute("INSERT INTO geodatabase (location, openet_id) VALUES (?, ?)", (key, value))
			self.connection.commit()
			cursor.close()
		except sqlite3.IntegrityError:
			pass # theoretically we've already cached it then, but it's weird that it tried to retrieve it if we checked beforehand?

	def check_gdb_cache(self, key):
		cursor = self.connection.cursor()
		cursor.execute("SELECT openet_id from geodatabase where location=:location_key", {"location_key": key})
		for record in cursor.fetchall():
			value = record[0]  # since we're only selection openet_id and the location key is unique, it'll be the first item in the only tuple returned
			break
		else:
			value = False  # return False if we didn't find something - None will be used for items that exist but are Null
		cursor.close()
		return value


import uuid
import tempfile
import shutil
import time
import functools
import logging

import requests

from .exceptions import BadRequestError, FileRetrievalError

STATUS_NONE = 0
STATUS_SUBMITTED = 1
STATUS_WAITING = 2
STATUS_AVAILABLE = 3
STATUS_DOWNLOADED = 4
STATUS_FAILED_OPENET = 5
STATUS_FAILED_CLIENT = 6


class Raster(object):
    """
        Internal object for managing raster exports - tracks current status, the remote URL and the local file path once
        it exists.
    """
    def __init__(self, request_result):
        self.status = STATUS_NONE
        self.params = {}
        self.remote_url = None
        self.local_file = None
        self.uuid = uuid.uuid4()

        self._request_result = request_result
        self._set_values()

    def _set_values(self):
        self.remote_url = self._request_result['bucket_url'][0]
        if "Queued" in self._request_result['status']:
            self.status = STATUS_SUBMITTED

    def download_file(self, retry_interval=20, max_wait=600):
        """
            Attempts to download a raster, assuming it's ready for download.
            Will make multiple attempts over a few minutes because sometimes it takes a while for the permissions
            to propagate, so the first few responses may give a 403 error. We then have a timeout (max_wait) where
            if we exceed that value, we exit anyway without downloading.

            Downloads the file to a tempfile path - the user may move the file after that if they
            wish.
        :param retry_interval: time in seconds between repeated attempts
        :param max_wait: How long, in seconds should we wait for the correct permissions before stopping attempts to download.
        :return:
        """
        # adapted from https://stackoverflow.com/a/39217788/587938
        local_filename = self.remote_url.split('/')[-1]
        local_file_path = tempfile.mktemp(local_filename)
        # NOTE the stream=True parameter below
        wait_time = 0
        while self.status == STATUS_AVAILABLE and wait_time < max_wait:  # keep trying - it can take time for the permissions to work out
            with requests.get(self.remote_url, stream=True) as r:
                r.raw.read = functools.partial(r.raw.read, decode_content=True)  # we need this to stream gzipped data correctly and decode it
                if r.status_code == 200 and r.headers["content-type"] == "image/tiff":  # if it's not a tiff, then we might be getting a text error back
                    with open(local_file_path, 'wb') as f:
                        shutil.copyfileobj(r.raw, f)

                    self.local_file = local_file_path
                    self.status = STATUS_DOWNLOADED
                    logging.info(f"Retrieved {self.local_file}")

                if r.status_code not in (200, 403):
                    # we skip 403 as well, because we're going to temporarily get 403s until permissions are set regardless.
                    raise FileRetrievalError(f"Couldn't retrieve {self.remote_url}. Received HTTP Status code {r.status_code}.")

                if not self.status == STATUS_DOWNLOADED:
                    logging.info(f"not yet available - trying again in {retry_interval}")
                    time.sleep(retry_interval)
                    wait_time += retry_interval


class RasterManager(object):
    """
        The manager that becomes the .raster attribute on the OpenETClient object.
        Handles submitting raster export requests and polling for completed exports.

        As constructed, could slow down if it handles many thousands of raster exports, or
        if the all_files OpenET endpoint displays lots of files as options.

        Generally speaking, you won't create this object yourself, but you can set
        client.raster.wait_interval to the length of time, in seconds, that the manager
        should wait between polling the all_files endpoint for new exports when waiting for new
        rasters.
    """

    client = None
    wait_interval = 30

    def __init__(self, client):
        self.client = client
        self.registry = {}

    def export(self, params=None, synchronous=False, public=True, transform=False):
        """
            Handles the raster/export endpoint for OpenET. Optionally waits for the raster to be exported
            and downloaded before proceeding. See documentation examples for usage details.

        :param params: A dictionary of arguments with keys matching the raster/export endpoints parameters
                        and values matching the requirements for the values of those keys. If the "geometry"
                        key of this value is a GEOS or an OGR object with a .coords attribute, then
                        the correct geometry string to send to the API will be composed for you. Otherwise,
                        it is your responsibility to match the values with the API's requirements.
        :param synchronous: Whether or not to wait for the raster to export and be downloaded before
                            exiting this function and proceeding
        :param public:  Whether or not to make the raster public - at this point, keeping this as True is
                        required for all the features of this package to work, but if you just want to use the
                        package to submit a bunch of raster jobs, but not to *download* those rasters, then
                        you may set this to False.
        :param transform: If a GEOS or OGR object with :code:`transform` and :code:`coords` attributes is passed in as part
                        of params["geography"],
                        such as an object from GeoDjango, then you can optionally specify that this function
                        handle a simple transformation to WGS84 (EPSG 4326) for you. For more complicated transformations
                        or datum transformations, it may be best to handle the transformation before running this function.
                        Setting it to False doesn't control usage of a GEOS/OGR object is used, only if its coordinates
                        are transformed first.
        :return: Raster object - when synchronous, the local_file attribute will
                        have the path to the downloaded raster on disk - otherwise it
                        will have the status of the raster
        """
        endpoint = "raster/export"
        params = {} if params is None else params

        if "filename_suffix" in params and not "public" in params["filename_suffix"] and public is True:
            params["filename_suffix"] += "_public"
        elif "filename_suffix" not in params and public is True:
            params["filename_suffix"] = "_public"

        # check that if they passed in an object for the geometry, see if we can get the correct coordinate strings.
        if hasattr(params["geometry"], "transform") and hasattr(params["geometry"], "coords"):
            geom = params["geometry"]
            if transform:
                geom = geom.transform(4326, clone=True)
            params["geometry"] = str(geom.coords).replace("(", "").replace(")", "").replace(" ", "")[:-1]
        #
        # it checks if we have the ability to reproject the geometry data passed in and if we can get a coordinate string from it,
        # such as from OGR and GEOS objects in GeoDjango. If it has those abilities, then it attempts to reproject, get the coordinates, and then
        # stringifies them and removes parens, spaces, and the trailing comma.

        result = self.client.send_request(endpoint, **params)

        if result.status_code not in (200, 201, 301) or "ERROR" in result.json():
            raise BadRequestError(f"OpenET API returned status code {result.status_code} with reason {result.reason} and message {result.text}")

        raster = Raster(result.json())
        self.registry[raster.uuid] = raster

        if synchronous:
            self.wait_for_rasters(raster.uuid)

        return raster

    @property
    def queued_rasters(self):
        """
            Which rasters are we still waiting for?
        :return:
        """

        return [raster for raster in self.registry.values() if raster.status < STATUS_DOWNLOADED ]

    @property
    def available_rasters(self):
        """
            Which rasters have we marked as ready to download, but haven't yet been retrieved?
        :return:
        """
        return [raster for raster in self.registry.values() if raster.status == STATUS_AVAILABLE]

    @property
    def downloaded_raster_paths(self):
        return [raster.local_file for raster in self.registry.values() if raster.status == STATUS_DOWNLOADED]

    def download_available_rasters(self):
        """
            Attempts to download all available rasters individually
        :return:
        """
        rasters = self.available_rasters

        for raster in rasters:
            raster.download_file()

    def wait_for_rasters(self, uuid=None, max_time=86400):
        """
            When we want to just wait until the rasters are ready, we call this method, which polls
            the all_files endpoint at set intervals and checks which rasters are done.

            Then updates the statuses on individual rasters and calls the download functions
            for the individual rasters and time some are ready to download, but won't exit until
            all rasters are available and have had at least one download attempt.
        :param uuid:
        :param max_time: Maximum time in seconds to wait for all rasters to complete - defaults to 86400 (a day)
        :return:
        """

        if uuid is None:
            rasters = self.queued_rasters
        else:
            rasters = [self.registry[uuid],]

        wait_time = 0
        while len(rasters) > 0 and wait_time < max_time:
            time.sleep(self.wait_interval)
            wait_time += self.wait_interval  # we'll have some error in this approach because we won't account for the time we spend processing things. We could just check how long it's been since we started waiting too

            self.check_statuses(rasters)

            self.download_available_rasters()
            rasters = [raster for raster in rasters if raster.status < STATUS_AVAILABLE]

    def check_statuses(self, rasters=None):
        """
            Updates the status information on each raster only - does not attempt to download them.
        :param rasters: the list of rasters to update the status of - if not provided, defaults to all rasters that
                        are queued and not downloaded
        :return: None
        """
        endpoint = "raster/export/all_files"

        if rasters is None:
            rasters = self.queued_rasters

        results = self.client.send_request(endpoint)
        for raster in rasters:
            if raster.remote_url in results.json()["rasters"]:
                raster.status = STATUS_AVAILABLE




import logging

import requests

logging.basicConfig()

from . import raster
from .raster import RasterManager
from .geodatabase import Geodatabase
from .exceptions import AuthenticationError, RateLimitError
from .cache import Cacher


class OpenETClient(object):
    token = None
    _base_url = "https://openet.dri.edu/"
    _validate_ssl = False

    def __init__(self, token=None):
        self.token = token
        self.raster = RasterManager(client=self)
        self.geodatabase = Geodatabase(client=self)
        self.cache = Cacher()
        self._last_request = None  # just for debugging

    def _check_token(self):
        if self.token is None:
            raise AuthenticationError("Token missing/undefined - you must set the value of your token before proceeding")

    def send_request(self, endpoint, method="get", disable_encoding=False, **kwargs):
        """
            Handles sending most requests to the API - they provide the endpoint and the args.
            Since the API is in the process of switching from GET to POST requests, we have logic that switches between
            those depending on the request method
        :param endpoint: The text path to the OpenET endpoint - e.g. raster/export - skip the base URL.
        :param method: "get" or "post" (case sensitive) - should match what the API supports for the endpoint
        :param kwargs: The arguments to send (via get or post) to the API
        :return: requests.Response object of the results.
        """

        self._check_token()

        requester = getattr(requests, method)
        send_kwargs = kwargs
        # they're not currently switching between post args and get args - it's just a get request that we POST instead...
        # send_kwargs = {}
        #if method == "get":
        #    send_kwargs["params"] = kwargs
        #    send_kwargs["data"] = None
        #elif method == "post":
        #    send_kwargs["data"] = kwargs
        #    send_kwargs["params"] = None

        url = self._base_url + endpoint
        logging.info(f"Connecting to {url}")
        logging.info(f"Sending params {kwargs}")
        logging.info(f"Sending token in header{self.token}")

        extra_kwargs = {}
        if self._validate_ssl != True:
            extra_kwargs['verify'] = False

        if disable_encoding:  # the API doesn't always like certain things URL-encoded, so don't
            send_kwargs = "&".join("%s=%s" % (k, v) for k, v in send_kwargs.items())

        result = requester(url, headers={"Authorization": self.token}, params=send_kwargs, **extra_kwargs)
        self._last_request = result

        if result.status_code == 500 and "reached your maximum rate limit" in result.json()["description"]:
            raise RateLimitError("Server indicates we've reached our rate limit - try increasing the wait time between requests")

        return result



import setuptools
import datetime

try:
    from openet_client import __version__ as version, __author__ as author
except ImportError:
    version = datetime.datetime.utcnow().strftime("%Y.%m.%d")
    author = "nickrsan"

if __name__ == "__main__":
    setuptools.setup(
        name="openet_client",
        version=version,
        packages=setuptools.find_packages(exclude=("tests",)),
        description="Client for the OpenET web API with useful wrappers to support common workflows and needs with the API",
        long_description="See README at https://github.com/water3d/openet/ for more details. Make sure to install package extras (e.g. pip install 'openet-client[spatial]') for spatial data support in the geodatabase API. It depends on Geopandas, which depends on fiona - this can be challenging to get set up correctly, so it's worth checking the documentation for those projects to install on your system, or use a conda environment and install the conda geopandas package.",
        license="MIT",
        author=author,
        author_email="nsantos5@ucmerced.edu",
        url='https://github.com/water3d/openet/',
        install_requires=["requests"],
        extras_requires={"spatial": ["geopandas"]},
        include_package_data=True,
    )


import tempfile
import time
import logging
from collections import OrderedDict

from .exceptions import RateLimitError

try:
	import fiona  # try importing fiona directly, because otherwise geopandas defers errors to later on when it actually needs to use it
	import geopandas
	GEOPANDAS_AVAILABLE = True
except ImportError:
	GEOPANDAS_AVAILABLE = False
	logging.warning("Can't load fiona or geopandas - will not be able to undertake spatial operations")

import pandas

MAX_FEATURE_IDS_LIST_LENGTH = 40
RATE_LIMIT = 5000  # ms

FEATURE_TYPE_GEOPANDAS = "geopandas"
FEATURE_TYPE_GEOJSON = "geojson"
FEATURE_TYPE_ARCPY = "arcpy"
FEATURE_TYPE_SHAPEPLY = "shapely"


def get_coords_shapely(geometry):
	coords = geometry.centroid.coords
	return (coords.x, coords.y)


def get_coords_arcpy(geometry):
	centroid = geometry.centroid.projectAs(4326)
	return (centroid.X, centroid.Y)


class Geodatabase(object):

	def __init__(self, client):
		self.client = client

	def get_et_for_features(self,
							params,
							features,
							feature_type,
							output_field=None,
							geometry_field="geometry",
							endpoint="timeseries/features/stats/annual",
							wait_time=RATE_LIMIT,
							batch_size=MAX_FEATURE_IDS_LIST_LENGTH,
							return_type="joined",
							join_type="outer"):
		"""
			Takes one of multiple data formats (user specified, we're not inspecting it - options are
			geopandas, geojson) and gets its
			coordinate values, then gets the field IDs in OpenET for the coordinate pair, retrieves the ET data
			and returns it as a geopandas data frame with the results in the specified output_field
		:param params:
		:param features:
		:param endpoint: which features endpoint should it use?
		:param return_type: How should we return the data? Options are "raw" to return just the JSON from OpenET,
							"list" to return a list of dictionaries with the OpenET data, "pandas" to return a pandas
							 data frame of the results, or "joined" to return the
							data joined back to the input data. "joined" is the default.
		:param join_type: When merging results back in, what type of join should we use? Defaults to "outer" so that
						records are retained even if no results come back for them. This is also useful behavior when
						we have multiple timeseries records, such as for monthly results, but it can duplicate input
						records (not always desirable). To change the behavior, change this to any value supported
						by pandas.merge or change the return_type so no join occurs.
		:return:
		"""

		if GEOPANDAS_AVAILABLE is False:
			# we'll check it this way because that way we can let people who don't want to get a working fiona/geopandas environment
			# use the application without it confusingly failing on them at runtime.
			raise EnvironmentError("Fiona or Geopandas is unavailable - check that Fiona and Geopandas are both installed and that importing Fiona works - cannot proceed without a working installation with fiona and geopandas")

		if endpoint.startswith("timeseries/"):  # strip it off the front if they included it
			endpoint.replace("timeseries/", "")

		if output_field is None and return_type == "joined":
			raise ValueError("Must specify value for output_field when return_type is 'joined'")

		if return_type not in ("joined", "pandas", "list", "raw"):
			raise ValueError("return_type must be one of ('joined', 'list', 'raw', 'pandas')")

		if feature_type not in (FEATURE_TYPE_GEOPANDAS, FEATURE_TYPE_GEOJSON):
			raise ValueError(f"Feature type must be in ({FEATURE_TYPE_GEOPANDAS}, {FEATURE_TYPE_GEOJSON}) to get geometries and retrieve ET. CHeck that the feature_type parameter is specified correctly")

		if feature_type == FEATURE_TYPE_GEOJSON:
			features = geopandas.GeoDataFrame.from_features(features)

		features_wgs = features.to_crs(4326)
		features_wgs.loc[:, "centroid_geom"] = features_wgs[geometry_field].centroid
		def set_centroid(row):
			"""
				There's a better way to do this, but my Pandas-fu is failing me right now.
				Make a function to set the centroid as text elementwise
			:param row:
			:return:
			"""
			# get the values as a string, but truncate it to 7 places for precision so that we can more reliably cache it
			row["centroid"] = f'{round(row["centroid_geom"].x, 7)} {round(row["centroid_geom"].y, 7)}'
			return row
		features_wgs = features_wgs.apply(set_centroid, axis=1)
		features_wgs = features_wgs.drop(columns=["centroid_geom"])  # drop it so it doesn't create output problems later

		# we're going to have to get the feature IDs one by one if we want a reliable mapping of polygons to openET features
		# which isn't ideal and we'll want to rate limit it to make sure we don't abuse the API too heavily
		# we'll probably also want to do some form of caching or saving the feature IDs to the geopandas dfs so that
		# we don't have to go back and get it again if we already got it.

		# only get the feature IDs if they aren't already there to save time and
		# avoid a column naming conflict if they run the same data through multiple times
		if not "openet_feature_id" in list(features_wgs.columns):
			openet_feature_ids = self.get_feature_ids(features_wgs, field="centroid")
			#temp_feature_outputs = tempfile.mktemp(suffix=".csv", prefix="openet_client")
			#openet_feature_ids.to_csv(temp_feature_outputs)

			features_wgs = features_wgs.merge(openet_feature_ids, on="centroid")

		feature_ids = features_wgs["openet_feature_id"].tolist()

		df_length = len(feature_ids)
		start = 0
		end = min(batch_size, df_length)
		results = []

		original_batch_size = batch_size
		slow_batch_count = 0
		while start < df_length:
			partial_list = [feat for feat in feature_ids[start:end] if feat is not None]  # remove the null values and filter to the batch size
			params["field_ids"] = str(partial_list).replace(" ", "").replace("\'", '"')  # what's weird is we basically have to send this as a python list, so we need to stringify it first so requests doesn't process it

			try:
				response = self.client.send_request(endpoint, method="post", disable_encoding=True, **params)
			except RateLimitError as e:
				# if it gets interrupted save the data we currently have to the exception then raise it up
				raise RateLimitError(str(e) + ". The retrieved data is available as an attribute '.data' on this exception, but is incomplete.", data=self._process_results(results, return_type, output_field, features_wgs, join_type))

			if not response.status_code == 500:
				results.extend(response.json())
			else:
				logging.warning(f"Error retrieving ET for one or more fields. Request sent was {response.url}. Got response {response.text}")
				if batch_size == original_batch_size:  # if we're not already there, switch to slow batch mode so we go through it one by one now
					batch_size = 1
					end = start + 1
					continue  # go back through the last batch one by one so we make sure we get as many as possible
				# if we are already in slow batch mode, then basically, this record gets skipped

			time.sleep(wait_time / 1000)

			start += batch_size

			if batch_size != original_batch_size:  # if we're in slow batch mode
				slow_batch_count += 1  # count how many we've done
				if slow_batch_count == original_batch_size:  # until we get back to where we would have been in the first place
					batch_size = original_batch_size  # then increase the batch size again to *try* to get a larger set for the next group

			end += batch_size
			end = min(end, df_length)  # we'll only check end because we won't enter the next iteration if start < df_length

		return self._process_results(results, return_type, output_field, features_wgs, join_type)

	def _process_results(self, results, return_type, output_field, features_wgs, join_type):
		if return_type == "raw":
			return results
	
		# openet_output_field_name = "data_value" if "aggregation" not in params else params["aggregation"]

		# figure out which keys are there using the first result - there should only be one, but this lets us make sure we get anything
		output_field_keys = list(set(results[0].keys()).intersection(set(["data_value", "sum", "mean", "min", "max", "median"])))

		results_reformed = results
		for item in results:
			if output_field:
				if len(output_field_keys) == 1:  # there should only be one
					item[output_field] = item[output_field_keys[0]]
					del item[output_field_keys[0]]
				else:
					for key in output_field_keys:
						item[output_field + "_" + key] = item[key]
						del item[key]
			item["openet_feature_id"] = item["feature_unique_id"]
			del item["feature_unique_id"]
		# we had used a list comprehension, but we wanted to keep all the other keys in the dict. Preserving for now, but can remove later
		#results_reformed = [{output_field: item[openet_output_field_name], "openet_feature_id": item["feature_unique_id", ]}
		#				for item in results]
		if return_type == "list":
			return results_reformed

		results_df = pandas.DataFrame(results_reformed)
		if return_type == "pandas":
			return results_df

		final = features_wgs.merge(results_df, on="openet_feature_id", how=join_type)
		return final

	def get_feature_ids(self, features, field=None, wait_time=RATE_LIMIT):
		"""
			An internal method used to get a list of coordinate pairs and return the feature ID. Values come back as a dictionary
			where the input item in the list (coordinate pair shown as DD Longitude space DD latitude)
			is a dictionary key and the value is the OpenET featureID
		:param features:
		:param field: when field is defined, features will be a pandas data frame with a field that has the coordinate values to use.
						In that case, results will be joined back to the data frame as the field openet_feature_id.
		:param wait_time: how long in ms should we wait between subsequent requests?
		:return:
		"""
		if field and not isinstance(features, pandas.DataFrame):
			raise ValueError("A field name was provided, but `features` are not a Pandas DataFrame. Must be a DataFrame to proceed, or a field name should not be provided")

		if field:
			inputs = features[field]
		else:
			inputs = features

		outputs = OrderedDict()
		for item in inputs:
			# check the cache first - we might not need an API request for their field ID
			cached_value = self.client.cache.check_gdb_cache(key=item)
			if cached_value is False:  # False indicates no records, None indicates it's there and Null
				params = {"coordinates": item, "spatial_join_type": "intersect", "override": "False"}
				results = self.feature_ids_list(params)
				results_dict = results.json()
				if "feature_unique_ids" in results_dict:
					ids = results_dict["feature_unique_ids"]
				else:
					logging.error(f"Unable to retrieve field ID. Server returned {results_dict}")
					raise ValueError(f"Unable to retrieve field ID. Server returned {results_dict}")

				if len(ids) > 0:
					outputs[item] = ids[0]
				else:
					outputs[item] = None

				# save the returned value in our cache so we don't make another roundtrip if we run these
				# same values through in the future
				self.client.cache.cache_gdb_item(key=item, value=outputs[item])
				time.sleep(wait_time / 1000)
			else:
				outputs[item] = cached_value
				# no need to sleep when we check out own cache!

		if field:
			out_df = pandas.DataFrame({field: outputs.keys(), "openet_feature_id": outputs.values()})
			out_df.set_index(keys=field)
			return out_df
		else:
			return outputs

	def feature_ids_list(self, params=None):
		"""
			The base OpenET Method - sends the supplied params to metadata/openet/region_of_interest/feature_ids_list
			and returns the requests.Response object
		:param params:
		:return:
		"""
		endpoint = "metadata/openet/region_of_interest/feature_ids_list"
		if params is None:
			params = {}
		results = self.client.send_request(endpoint, method="post", **params)
		return results



class DataProcessingError(RuntimeError):
    def __init__(self, text, data=None):
        self.text = text
        self.data = data

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text


class AuthenticationError(PermissionError):
    pass


class BadRequestError(ValueError):
    pass


class RateLimitError(DataProcessingError):
    pass

class FileRetrievalError(RuntimeError):
    pass